import os

import testresources
from testscenarios import generate_scenarios


def load_tests(loader, tests, pattern):
    this_dir = os.path.dirname(__file__)
    mytests = loader.discover(start_dir=this_dir, pattern=pattern)
    result = testresources.OptimisingTestSuite()
    result.addTests(generate_scenarios(mytests))
    result.addTests(generate_scenarios(tests))
    return result



# Copyright (c) 2021-2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import requests
import testtools

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import gem
from soufi.testing import base


class TestGemFinder(base.TestCase):
    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        return gem.GemFinder(name, version, SourceType.gem)

    def test_get_source_url(self):
        finder = self.make_finder()
        url = f'{gem.GEM_DOWNLOADS}{finder.name}-{finder.version}.gem'

        head = self.patch_head_with_response(requests.codes.ok)
        found_url = finder.get_source_url()
        self.assertEqual(found_url, url)
        head.assert_called_once_with(url, timeout=30)

    def test_get_source_info_raises_when_response_fails(self):
        self.patch_head_with_response(requests.codes.not_found)
        finder = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.get_source_url()

    def test_find(self):
        url = self.factory.make_url()
        finder = self.make_finder()
        self.patch(finder, 'get_source_url').return_value = url

        disc_source = finder.find()
        self.assertIsInstance(disc_source, gem.GemDiscoveredSource)
        self.assertEqual([url], disc_source.urls)


class TestGemDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None):
        if url is None:
            url = self.factory.make_url()
        return gem.GemDiscoveredSource([url])

    def test_repr(self):
        url = self.factory.make_url()
        gem_discovered_source = self.make_discovered_source(url)
        self.assertEqual(url, repr(gem_discovered_source))

    def test_make_archive(self):
        gds = self.make_discovered_source()
        self.assertEqual(gds.make_archive, gds.remote_url_is_archive)



# Copyright (c) 2021-2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import pathlib
import tarfile
import tempfile
from unittest import mock

import fixtures
import requests
import testtools
from testtools.matchers import Equals
from testtools.matchers._basic import SameMembers

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import debian
from soufi.testing import base


class TestDebianFinder(base.TestCase):
    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        return debian.DebianFinder(name, version, SourceType.os)

    def make_source_info(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string()
        if version is None:
            version = self.factory.make_string()
        return dict(name=name, version=version)

    def test_get_source_info(self):
        df = self.make_finder()
        source_version = self.factory.make_string()
        source_name = self.factory.make_string()

        data = dict(
            result=[
                {
                    'binary_version': self.factory.make_string(),
                    'source': self.factory.make_string(),
                    'version': self.factory.make_string(),
                    'name': self.factory.make_string(),
                },
                {
                    'binary_version': df.version,
                    'source': source_name,
                    'version': source_version,
                    'name': df.name,
                },
            ]
        )
        self.patch_get_with_response(requests.codes.ok, json=data)
        source_info = df.get_source_info()
        expected = dict(name=source_name, version=source_version)
        self.assertDictEqual(expected, source_info)

    def test_get_source_info_raises_when_response_fails(self):
        self.patch_get_with_response(requests.codes.not_found)
        df = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            df.get_source_info()

    def test_get_source_info_raises_when_cant_find_version(self):
        df = self.make_finder()
        data = dict(
            result=[
                {
                    'binary_version': self.factory.make_string(),
                    'source': self.factory.make_string(),
                    'version': self.factory.make_string(),
                    'name': self.factory.make_string(),
                },
            ]
        )
        self.patch_get_with_response(requests.codes.ok, json=data)
        with testtools.ExpectedException(exceptions.SourceNotFound):
            df.get_source_info()

    def test_get_hashes(self):
        df = self.make_finder()
        hashes = [self.factory.make_digest() for _ in range(4)]
        data = dict(result=[{'hash': hash} for hash in hashes])
        get = self.patch_get_with_response(requests.codes.ok, json=data)
        source_info = self.make_source_info()
        self.assertEqual(hashes, df.get_hashes(source_info))
        get.assert_called_once_with(
            f"{debian.SNAPSHOT_API}mr/package/"
            f"{source_info['name']}/{source_info['version']}/srcfiles",
            timeout=df.timeout,
        )

    def test_get_hashes_raises_for_requests_error(self):
        self.patch_get_with_response(requests.codes.bad_request)
        df = self.make_finder()
        source_info = self.make_source_info()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            df.get_hashes(source_info)

    def test_get_hashes_raises_for_response_error(self):
        df = self.make_finder()
        source_info = self.make_source_info()
        self.patch_get_with_response(requests.codes.ok, json=[])
        with testtools.ExpectedException(exceptions.SourceNotFound):
            df.get_hashes(source_info)

    def test_get_urls(self):
        df = self.make_finder()
        hashes = [self.factory.make_digest() for _ in range(2)]
        name = self.factory.make_string()
        data = dict(
            result=[dict(name=name), dict(name=self.factory.make_string())]
        )
        get = self.patch_get_with_response(requests.codes.ok, json=data)
        expected = [
            (name, f"{debian.SNAPSHOT_API}file/{hash}") for hash in hashes
        ]
        self.assertEqual(expected, df.get_urls(hashes))
        calls = [
            mock.call(
                f"{debian.SNAPSHOT_API}mr/file/{hash}/info",
                timeout=df.timeout,
            )
            for hash in hashes
        ]
        self.assertEqual(calls, get.call_args_list)

    def test_get_urls_raises_for_requests_error(self):
        df = self.make_finder()
        self.patch_get_with_response(requests.codes.bad_request)
        with testtools.ExpectedException(exceptions.DownloadError):
            df.get_urls(['foo'])

    def test_find(self):
        source_info = self.make_source_info()
        get_source_info = self.patch(debian.DebianFinder, 'get_source_info')
        get_source_info.return_value = source_info
        hashes = [self.factory.make_digest for _ in range(4)]
        urls = [self.factory.make_url for _ in range(4)]
        url_pairs = [(self.factory.make_string(), url) for url in urls]
        get_hashes = self.patch(debian.DebianFinder, 'get_hashes')
        get_hashes.return_value = hashes
        get_urls = self.patch(debian.DebianFinder, 'get_urls')
        get_urls.return_value = url_pairs
        df = self.make_finder()

        disc_source = df.find()

        get_source_info.assert_called_once()
        get_hashes.assert_called_once_with(source_info)
        get_urls.assert_called_once_with(hashes)
        names, urls = zip(*url_pairs)
        self.assertEqual(urls, disc_source.urls)
        self.assertEqual(names, disc_source.names)


class TestDebianDiscoveredSource(base.TestCase):
    def make_debian_discovered_source(self, url_count=3):
        urls = [self.factory.make_url() for _ in range(url_count)]
        url_pairs = [(self.factory.make_string(), url) for url in urls]
        return url_pairs, debian.DebianDiscoveredSource(url_pairs)

    def test_repr(self):
        url_pairs, dds = self.make_debian_discovered_source()
        expected = "\n".join([f"{name}: {url}" for name, url in url_pairs])
        self.assertEqual(expected, repr(dds))

    def test_populate_archive(self):
        tmpdir = self.useFixture(fixtures.TempDir()).path

        # Make a file to put in the tar:
        content = self.factory.make_bytes('content')
        _, fake_file = tempfile.mkstemp(dir=tmpdir)
        with open(fake_file, 'wb') as fake_file_fd:
            fake_file_fd.write(content)

        # Patch out download_file to return the fake file:
        url_pairs, dds = self.make_debian_discovered_source(url_count=1)
        download_file = self.patch(dds, 'download_file')
        download_file.return_value = pathlib.Path(fake_file)

        # Make the tar file and populate it:
        _, tar_file_name = tempfile.mkstemp(dir=tmpdir)
        with tarfile.open(name=tar_file_name, mode='w') as tar:
            dds.populate_archive(tmpdir, tar)

        # Test that the tar contains the file and the content.
        expected_file_name, _ = url_pairs[0]
        with tarfile.open(name=tar_file_name, mode='r') as tar:
            self.expectThat(tar.getnames(), SameMembers([expected_file_name]))
            self.expectThat(
                tar.extractfile(expected_file_name).read(), Equals(content)
            )



# Copyright (c) 2021-2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import requests
import testtools

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import java
from soufi.testing import base


class TestJavaFinder(base.TestCase):
    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        return java.JavaFinder(name, version, SourceType.java)

    def test_get_source_url(self):
        finder = self.make_finder()
        group = self.factory.make_string()
        url = self.factory.make_url()
        data = dict(response=dict(docs=[dict(g=group)]))

        get = self.patch_get_with_response(requests.codes.ok, json=data)
        head = self.patch_head_with_response(requests.codes.ok)
        head.return_value.url = url
        found_url = finder.get_source_url()
        self.assertEqual(found_url, url)
        expected = dict(
            q=f'a:{finder.name} v:{finder.version} l:sources', rows=1
        )
        get.assert_called_once_with(
            java.MAVEN_SEARCH_URL, params=expected, timeout=finder.timeout
        )
        expected = dict(
            filepath=f'{group}/{finder.name}/{finder.version}/'
            f'{finder.name}-{finder.version}-sources.jar'
        )
        head.assert_called_once_with(
            java.MAVEN_REPO_URL,
            params=expected,
            allow_redirects=True,
            timeout=finder.timeout,
        )

    def test_get_source_info_raises_when_get_response_fails(self):
        self.patch_get_with_response(requests.codes.bad)
        finder = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.get_source_url()

    def test_get_source_info_raises_when_head_response_fails(self):
        group = self.factory.make_string()
        data = dict(response=dict(docs=[dict(g=group)]))
        self.patch_get_with_response(requests.codes.ok, json=data)
        self.patch_head_with_response(requests.codes.not_found)
        finder = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.get_source_url()

    def test_get_source_info_raises_when_head_response_times_out(self):
        group = self.factory.make_string()
        data = dict(response=dict(docs=[dict(g=group)]))
        self.patch_get_with_response(requests.codes.ok, json=data)
        head = self.patch(requests, 'head')
        head.side_effect = requests.exceptions.Timeout
        finder = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.get_source_url()

    def test_find(self):
        url = self.factory.make_url()
        finder = self.make_finder()
        self.patch(finder, 'get_source_url').return_value = url

        disc_source = finder.find()
        self.assertIsInstance(disc_source, java.JavaDiscoveredSource)
        self.assertEqual([url], disc_source.urls)


class TestNPMDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None):
        if url is None:
            url = self.factory.make_url()
        return java.JavaDiscoveredSource([url])

    def test_repr(self):
        url = self.factory.make_url()
        jds = self.make_discovered_source(url)
        self.assertEqual(url, repr(jds))

    def test_make_archive(self):
        jds = self.make_discovered_source()
        self.assertEqual(jds.make_archive, jds.remote_url_is_archive)



# Copyright (c) 2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import requests

from soufi.finder import SourceType
from soufi.finders import almalinux, yum
from soufi.testing import base


class BaseAlmaTest(base.TestCase):
    def make_finder(self, name=None, version=None, **kwargs):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        if 'source_repos' not in kwargs:
            kwargs['source_repos'] = ['']
        if 'binary_repos' not in kwargs:
            kwargs['binary_repos'] = ['']
        return almalinux.AlmaLinuxFinder(
            name, version, SourceType.os, **kwargs
        )

    def make_href(self, text):
        return f'<a href="{text}">{text}</a>'

    def make_top_page_content(self, versions):
        links = [self.make_href(v) for v in versions]
        return "\n".join(links)


class TestAlmaFinder(BaseAlmaTest):
    def test_find(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        self.patch(finder, 'get_source_url').return_value = url
        disc_source = finder.find()
        self.assertIsInstance(disc_source, yum.YumDiscoveredSource)
        self.assertEqual([url], disc_source.urls)

    def test__get_dirs(self):
        finder = self.make_finder()
        top_repos = ('1.0.123', '2.1.3456', 'bogus', '3.7.89-beta', '3')
        top_data = self.make_top_page_content(top_repos)
        get = self.patch_get_with_response(requests.codes.ok, top_data)
        result = list(finder._get_dirs())
        # Ensure that only the items we're interested in come back
        self.assertEqual(['2.1.3456', '1.0.123'], result)
        get.assert_called_once_with(almalinux.VAULT, timeout=30)

    def test__get_source_repos(self):
        finder = self.make_finder()
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.return_value = True
        result = list(finder.get_source_repos())
        expected = [
            f"{almalinux.VAULT}/{dir}/{subdir}/Source/"
            for dir in dirs
            for subdir in almalinux.DEFAULT_SEARCH
        ]
        self.assertEqual(expected, result)

    def test__get_binary_repos(self):
        finder = self.make_finder()
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.return_value = True
        result = list(finder.get_binary_repos())
        expected = [
            f"{almalinux.VAULT}/{dir}/{subdir}/x86_64/os/"
            for dir in dirs
            for subdir in almalinux.DEFAULT_SEARCH
        ]
        self.assertEqual(expected, result)



# Copyright (c) 2021-2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import itertools
from importlib import import_module
from unittest import mock
from unittest.mock import MagicMock

import requests
import testtools

from soufi.testing import factory


# TODO(nic): this needs to handle "unbalanced" parameter sets; if all
#  parameters are not provided with all scenarios, then the parameter lists
#  go out of sync with the inputs.  If you call this without properly
#  defining your input scenarios, you get to keep all the pieces.
# TODO(nic): give this to testscenarios, if we think it might be useful to them
def combine_scenarios(*scenarios):
    """Create a combinatorial matrix of all provided scenarios.

    Like `testscenarios.scenarios.multiply_scenarios`, but it stacks up all
    possible combinations of scenarios rather than creating a cross-product.

    :returns: A list of compound scenarios: the combination of all
        scenarios, with the names concatenated and the parameters of all
        scenarios stacked into lists.
    """
    result = []
    scenario_lists = list(map(list, scenarios))
    for i in range(1, len(*scenarios) + 1):
        for combination in itertools.combinations(*scenario_lists, i):
            names, parameters = zip(*combination)
            scenario_name = ",".join(names)
            scenario_parameters = {}
            for parameter in parameters:
                for k, v in parameter.items():
                    if k not in scenario_parameters:
                        scenario_parameters[k] = []
                    scenario_parameters[k].append(v)
            result.append((scenario_name, scenario_parameters))
    return result


class TestCase(testtools.TestCase):
    """Base class for all tests."""

    factory = factory.factory

    def patch(
        self, obj, attribute=None, value=mock.sentinel.unset
    ) -> MagicMock:
        """Patch `obj.attribute` with `value`.

        If `value` is unspecified, a new `MagicMock` will be created and
        patched-in instead. Its ``__name__`` attribute will be set to
        `attribute` or the ``__name__`` of the replaced object if `attribute`
        is not given.

        This is a thin customisation of `testtools.TestCase.patch`, so refer
        to that in case of doubt.

        :return: The patched-in object.
        """
        # If 'attribute' is None, assume 'obj' is a 'fully-qualified' object,
        # and assume that its __module__ is what we want to patch. For more
        # complex use cases, the two-parameter 'patch' will still need to
        # be used.
        if attribute is None:
            attribute = obj.__name__
            obj = import_module(obj.__module__)
        else:
            if not hasattr(obj, attribute):
                # Prevent footgunning.
                raise AttributeError(f"{attribute} does not exist on {obj}")
        if value is mock.sentinel.unset:
            value = MagicMock(__name__=attribute)
        super().patch(obj, attribute, value)
        return value

    def extend_side_effects(self, mock_obj, extra_value):
        """Cumulatively add side effects to an existing Mock.

        :param mock_obj: The Mock object
        :param extra_value: Any object which to add to the
            Mock.side_effect

        This exists because Mock implements side_effect as a list iterator,
        which means there's extra work than simply appending to an existing
        list of side_effects.
        """
        effects = list(mock_obj.side_effect)
        effects.append(extra_value)
        mock_obj.side_effect = effects

    def patch_get_with_response(self, response_code, data=None, json=None):
        """Patch `requests.get` with the provided values.

        :param response_code: A requests.codes value, to mimic an HTTP status
        :param data: A string-like object to mimic Response.content
        :param json: A dict or list, to mimic what Response.json() would return
        :return: The created MagicMock, to add side-effects, etc.
        """
        fake_response = mock.MagicMock()
        fake_response.return_value.status_code = response_code
        fake_response.return_value.content = data
        fake_response.return_value.json.return_value = json
        return self.patch(requests, 'get', fake_response)

    def patch_head_with_response(self, response_code):
        """Patch `requests.head` with the provided values.

        HEAD requests have an empty message body, so no data is accepted.

        :param response_code: A requests.codes value, to mimic an HTTP status
        :return: The created MagicMock, to add side-effects, etc.
        """
        fake_response = mock.MagicMock()
        fake_response.return_value.status_code = response_code
        return self.patch(requests, 'head', fake_response)



# Copyright (c) 2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import re

from lxml import html

import soufi.finders.yum as yum_finder
from soufi import finder

VAULT = "https://repo.almalinux.org/vault"
# AlmaLinux does their very best to be a faithful CentOS clone.  Up to and
# including this little gem.  See the CentosFinder for an explanation.
CURRENT = "https://repo.almalinux.org/almalinux"
# Rather than aimlessly poke around their CDN, use this hopefully-sensible
# list of search dirs when looking for repodata
DEFAULT_SEARCH = ('BaseOS', 'AppStream', 'extras', 'cloud', 'devel')


class AlmaLinuxFinder(yum_finder.YumFinder):
    """Find AlmaLinux source files.

    By default, Iterates over the index at https://repo.almalinux.org/vault/
    """

    distro = finder.Distro.almalinux.value

    def _get_dirs(self):
        """Get all the possible Vault dirs that could match."""
        content = self.get_url(VAULT).content
        tree = html.fromstring(content)
        # Ignore beta releases; we may want to make this a switchable behavior
        retval = tree.xpath("//a/text()[not(contains(.,'-beta'))]")
        # AlmaLinux Vault is fond of symlinking the current point release to a
        # directory with just the major version number, e.g., `6.10/`->`6/`.
        # This means that such directories are inherently unstable and their
        # contents are subject to change without notice, so we'll ignore
        # them in favour of the "full" names.
        dirs = [dir.rstrip('/') for dir in retval if re.match(r'\d+\.\d', dir)]

        # Walk the tree backwards, so that newer releases get searched first
        return reversed(dirs)

    def get_source_repos(self):
        """Determine which source search paths are valid.

        Spams the vault with HEAD requests and keeps the ones that hit.

        This is implemented as a generator so that the methods in the
        YumFinder base class can "load as it goes", rather than having to
        do a ton of discovery up-front that might end up being wasted.
        Absolutely everything is cached, so the relative overhead of having
        to run the generator over when re-walking the list of repos is minimal.
        """
        for dir in self._get_dirs():
            for subdir in DEFAULT_SEARCH:
                url = f"{VAULT.rstrip('/')}/{dir}/{subdir}/Source/"
                if self.test_url(url + "repodata/"):
                    yield url

    def get_binary_repos(self):
        """Determine which binary search paths are valid.

        Spams the vault with HEAD requests and keeps the ones that hit.

        This is also implemented as a generator.  See get_source_repos().
        """

        def _find_valid_repo_url(dir, subdir):
            vault_url = f"{VAULT.rstrip('/')}/{dir}/{subdir}/x86_64/"
            current_url = f"{CURRENT.rstrip('/')}/{dir}/{subdir}/x86_64/"
            for url in vault_url, current_url:
                if self.test_url(url + "os/repodata/"):
                    yield url + "os/"
                    break

        for dir in self._get_dirs():
            for subdir in DEFAULT_SEARCH:
                yield from _find_valid_repo_url(dir, subdir)



# Copyright (c) 2021-2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

from testtools.matchers import SameMembers

from soufi.finder import SourceType
from soufi.finders import rhel, yum
from soufi.testing import base


class BaseRHELTest(base.TestCase):
    def make_finder(self, name=None, version=None, **kwargs):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        if 'source_repos' not in kwargs:
            kwargs['source_repos'] = ['']
        if 'binary_repos' not in kwargs:
            kwargs['binary_repos'] = ['']
        return rhel.RHELFinder(name, version, SourceType.os, **kwargs)

    def test_find(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        self.patch(finder, 'get_source_url').return_value = url
        disc_source = finder.find()
        self.assertIsInstance(disc_source, yum.YumDiscoveredSource)
        self.assertEqual([url], disc_source.urls)

    def test_default_repos(self):
        name = self.factory.make_string('name')
        version = self.factory.make_string('version')
        finder = rhel.RHELFinder(name, version, SourceType.os)
        expected_source = [
            f"{rhel.DEFAULT_REPO}/{dir}/source/SRPMS"
            for dir in rhel.RHELFinder.default_search_dirs
        ]
        expected_binary = [
            f"{rhel.DEFAULT_REPO}/{dir}/os"
            for dir in rhel.RHELFinder.default_search_dirs
        ]
        self.assertThat(
            finder.get_source_repos(), SameMembers(expected_source)
        )
        self.assertThat(
            finder.get_binary_repos(), SameMembers(expected_binary)
        )



# Copyright (c) 2022 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import soufi.finders.yum as yum_finder
from soufi import finder

DEFAULT_REPO = "https://cdn-ubi.redhat.com/content/public/ubi/dist"


class RHELFinder(yum_finder.YumFinder):
    """Find Red Hat Enterprise Linux source files.

    By default, uses the public UBI index at https://cdn-ubi.redhat.com
    """

    distro = finder.Distro.rhel.value

    # UBI is kind of sprawling, and they are fairly idiosyncratic with where
    # they like to put their packages, tree-wise.  Once you find the proper
    # subtree, they are delightfully regular with the repo structure, however.
    # Rather than aimlessly poke around their CDN, we'll use this
    # hand-curated list of search paths.
    default_search_dirs = (
        'ubi8/8/x86_64/baseos',
        'ubi8/8/x86_64/appstream',
        'ubi8/8/x86_64/codeready-builder',
        'ubi/server/7/7Server/x86_64',
        'ubi/server/7/7Server/x86_64/extras',
        'ubi/server/7/7Server/x86_64/devtools/1',
        'ubi/server/7/7Server/x86_64/optional',
        'ubi/server/7/7Server/x86_64/rhscl/1',
        'ubi/atomic/7/7Server/x86_64',
    )

    def get_source_repos(self):
        for dir in self.default_search_dirs:
            yield f"{DEFAULT_REPO}/{dir}/source/SRPMS"

    def get_binary_repos(self):
        for dir in self.default_search_dirs:
            yield f"{DEFAULT_REPO}/{dir}/os"



# Copyright (c) 2021-2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

from unittest import mock

import requests
import testtools

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import python
from soufi.testing import base


class TestPythonFinder(base.TestCase):
    scenarios = [
        ('pypi', dict(index='pypi')),
        ('devpi', dict(index='devpi')),
    ]

    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        kwargs = dict(name=name, version=version, s_type=SourceType.python)
        if self.index == 'devpi':
            kwargs['pyindex'] = self.factory.make_url()
        return python.PythonFinder(**kwargs)

    def make_data(self, version, url, new_style=False):
        if self.index == 'devpi':
            result = dict()
            result['+links'] = []
            result['+links'].append(dict(href=url))
            result['+links'].append(dict(href=self.factory.make_url()))
            return dict(result=result)

        releases = {
            version: [
                dict(packagetype='foobar', url='foobar'),
                dict(packagetype='sdist', url=url),
            ]
        }
        urls = releases[version]

        if new_style:
            return dict(urls=urls)
        return dict(releases=releases)

    def get_requests_call_for_scenario(self, finder):
        if self.index == 'devpi':
            return mock.call(
                f"{finder.index}{finder.name}/{finder.version}",
                headers={'Accept': 'application/json'},
                timeout=30,
            )
        return mock.call(
            f"https://pypi.org/pypi/{finder.name}/{finder.version}/json",
            timeout=30,
        )

    def test_get_source_url(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        data = self.make_data(finder.version, url)

        get = self.patch_get_with_response(requests.codes.ok, json=data)
        found_url = finder.get_source_url()
        self.assertEqual(found_url, url)
        call = self.get_requests_call_for_scenario(finder)
        self.assertIn(call, get.call_args_list)

    def test_get_source_url_new_style(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        data = self.make_data(finder.version, url, new_style=True)

        get = self.patch_get_with_response(requests.codes.ok, json=data)
        found_url = finder.get_source_url()
        self.assertEqual(found_url, url)
        call = self.get_requests_call_for_scenario(finder)
        self.assertIn(call, get.call_args_list)

    def test_get_source_url_source_not_found(self):
        if self.index == 'devpi':
            # devpi returns data or 404.
            self.skipTest("Not applicable to devpi")
        finder = self.make_finder()
        data = self.make_data('badversion', 'foobar')

        self.patch_get_with_response(requests.codes.ok, json=data)
        self.assertRaises(exceptions.SourceNotFound, finder.get_source_url)

    def test_get_source_info_raises_when_response_fails(self):
        self.patch_get_with_response(requests.codes.not_found)
        finder = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.get_source_url()

    def test_find(self):
        url = self.factory.make_url()
        finder = self.make_finder()
        self.patch(finder, 'get_source_url').return_value = url

        disc_source = finder.find()
        self.assertIsInstance(disc_source, python.PythonDiscoveredSource)
        self.assertEqual([url], disc_source.urls)


class TestPythonDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None):
        if url is None:
            url = self.factory.make_url()
        return python.PythonDiscoveredSource([url])

    def test_repr(self):
        url = self.factory.make_url()
        pds = self.make_discovered_source(url)
        self.assertEqual(url, repr(pds))

    def test_make_archive(self):
        pds = self.make_discovered_source()
        self.assertEqual(pds.make_archive, pds.remote_url_is_archive)



# Copyright (c) 2022-2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import re
from typing import Iterable

from lxml import html

import soufi.finders.yum as yum_finder
from soufi import finder

VAULT = "https://vault.centos.org/centos/"
DEFAULT_SEARCH = ('BaseOS', 'os', 'updates', 'extras')
# Optimal search dirs considered a useful extended set to inspect in
# addition to the defaults.
OPTIMAL_SEARCH = ('AppStream', 'PowerTools', 'fasttrack')

# CentOS is adorable sometimes.  The current release on mirror.centos.org
# stores zero binary packages in the vault.  It does store all source
# packages, however.  So when looking up binary package repos,
# it is necessary to also check the main mirror when looking for the "current"
# releases.  Note that the connection is unencrypted, and no HTTPS endpoint
# is provided.
MIRROR = "http://mirror.centos.org/centos/"


class CentosFinder(yum_finder.YumFinder):
    """Find CentOS source files.

    By default, iterates over the index at https://vault.centos.org

    :param repos: An iterable of repo names in the [S]Packages directory of
        the source repo. E.g. 'os', 'extras', etc. If not specified,
        a default set of 'os', 'BaseOS', 'updates' and 'extras' are examined.
    :param optimal_repos: (bool) Override repos to select what is considered
        an optimal set to inspect. These are the above defaults, plus:
        'AppStream', 'PowerTools', 'fasttrack'
    """

    distro = finder.Distro.centos.value

    def __init__(
        self,
        *args,
        repos: Iterable[str] = None,
        optimal_repos: bool = False,
        source_repos: Iterable[str] = None,
        binary_repos: Iterable[str] = None,
        **kwargs,
    ):
        if not repos or optimal_repos is True:
            repos = DEFAULT_SEARCH
        self.repos = repos
        self.optimal_repos = optimal_repos
        super().__init__(
            *args,
            source_repos=source_repos,
            binary_repos=binary_repos,
            **kwargs,
        )

    def _get_dirs(self):
        """Get all the possible Vault dirs that could match."""
        content = self.get_url(VAULT).content
        tree = html.fromstring(content)
        dirs = tree.xpath('//td[@class="indexcolname"]/a/text()')
        # CentOS Vault is fond of symlinking the current point release to a
        # directory with just the major version number, e.g., `6.10/`->`6/`.
        # This means that such directories are inherently unstable and their
        # contents are subject to change without notice, so we'll ignore
        # them in favour of the "full" names.
        dirs = [dir.rstrip('/') for dir in dirs if re.match(r'\d+\.\d', dir)]

        # Walk the tree backwards, so that newer releases get searched first
        return reversed(dirs)

    def get_source_repos(self):
        """Determine which source search paths are valid.

        Spams the vault with HEAD requests and keeps the ones that hit.

        CentOS is *big*.  CentOS is old, and the layout of the historical
        repos in the vault is so sprawling, that on average it takes about
        2 minutes of wall-clock time just to figure out where all of the
        repos are, and then even more time to subsequently download them all.

        As such, this is implemented as a generator so that the methods in the
        YumFinder base class can "load as it goes", rather than having to
        do a ton of discovery up-front that might end up being wasted.
        Absolutely everything is cached, so the relative overhead of having
        to run the generator over when re-walking the list of repos is minimal.
        """
        for dir in self._get_dirs():
            for subdir in self.repos:
                url = f"{VAULT.rstrip('/')}/{dir}/{subdir}/Source/"
                if self.test_url(url + "repodata/"):
                    yield url
            if self.optimal_repos:
                for subdir in OPTIMAL_SEARCH:
                    url = f"{VAULT.rstrip('/')}/{dir}/{subdir}/Source/"
                    if self.test_url(url + "repodata/"):
                        yield url

    def get_binary_repos(self):
        """Determine which binary search paths are valid.

        Spams the vault with HEAD requests and keeps the ones that hit.

        This is also implemented as a generator.  See get_source_repos().
        """

        def _find_valid_repo_url(dir, subdir):
            vault_url = f"{VAULT.rstrip('/')}/{dir}/{subdir}/x86_64/"
            mirror_url = f"{MIRROR.rstrip('/')}/{dir}/{subdir}/x86_64/"
            for url in vault_url, mirror_url:
                if self.test_url(url + "os/repodata/"):
                    yield url + "os/"
                    break
                elif self.test_url(url + "repodata/"):
                    yield url
                    break

        for dir in self._get_dirs():
            for subdir in self.repos:
                yield from _find_valid_repo_url(dir, subdir)
            if self.optimal_repos:
                for subdir in OPTIMAL_SEARCH:
                    yield from _find_valid_repo_url(dir, subdir)



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.


from soufi import exceptions, finder

GEM_DOWNLOADS = 'https://rubygems.org/downloads/'


class GemFinder(finder.SourceFinder):
    """Find Gem files.

    Finds files at https://rubygems.org/downloads
    """

    distro = finder.SourceType.gem.value

    def _find(self):
        source_url = self.get_source_url()
        return GemDiscoveredSource([source_url], timeout=self.timeout)

    def get_source_url(self):
        url = f"{GEM_DOWNLOADS}{self.name}-{self.version}.gem"
        if not self.test_url(url):
            raise exceptions.SourceNotFound
        return url


class GemDiscoveredSource(finder.DiscoveredSource):
    """A discovered Gem package."""

    make_archive = finder.DiscoveredSource.remote_url_is_archive
    # NOTE(nic): `distro` and `archive_extension` are the same,
    # so auto-output mode will name these things `.gem.gem`.  The
    # recommended workaround is to not use auto-output mode :-)
    archive_extension = '.gem'

    def populate_archive(self, *args, **kwargs):  # pragma: no cover
        # Required by the base class but Gems are already tarballs so
        # nothing to do.
        pass

    def __repr__(self):
        return self.urls[0]



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import pathlib
import shutil
import tarfile
import tempfile
from io import BytesIO

import fixtures
import requests
import testtools
from testtools.matchers import DirExists, Equals, FileContains, Not
from testtools.matchers._basic import SameMembers

from soufi import exceptions
from soufi.finder import DiscoveredSource, SourceFinder, SourceType
from soufi.testing import base


class TestSourceFinderBase(base.TestCase):
    class TestFinder(SourceFinder):
        def distro(self):
            pass

        def _find(self):
            pass

    def test_stores_init_params(self):
        name = self.factory.make_string()
        version = self.factory.make_string()
        s_type = self.factory.pick_enum(SourceType)
        sf = self.TestFinder(name, version, s_type)
        self.expectThat(sf.name, Equals(name))
        self.expectThat(sf.version, Equals(version))
        self.expectThat(sf.s_type, Equals(s_type))

    def test_requires_distro_property(self):
        class TestFinder(SourceFinder):
            def find(self):
                pass

        name = self.factory.make_string()
        version = self.factory.make_string()
        s_type = self.factory.pick_enum(SourceType)
        with testtools.ExpectedException(TypeError) as e:
            TestFinder(name, version, s_type)
            self.assertIn('abstract methods distro', str(e))

    def test_requires_find_method(self):
        class TestFinder(SourceFinder):
            def distro(self):
                pass

        name = self.factory.make_string()
        version = self.factory.make_string()
        s_type = self.factory.pick_enum(SourceType)
        with testtools.ExpectedException(TypeError) as e:
            TestFinder(name, version, s_type)
            self.assertIn('abstract methods find', str(e))

    def test_find_overrides_constructor_values(self):
        name = self.factory.random_choice([None, self.factory.make_string()])
        version = self.factory.random_choice(
            [None, self.factory.make_string()]
        )
        s_type = self.factory.random_choice(
            [None, self.factory.pick_enum(SourceType)]
        )
        name2 = self.factory.make_string()
        version2 = self.factory.make_string()
        s_type2 = self.factory.pick_enum(SourceType)
        sf = self.TestFinder(name, version, s_type)
        sf.find(name2, version2, s_type2)

        self.expectThat(sf.name, Equals(name2))
        self.expectThat(sf.version, Equals(version2))
        self.expectThat(sf.s_type, Equals(s_type2))

    def test_error_if_missing_find_value(self):
        for values in [
            (None, self.factory.make_string(), self.factory.make_string()),
            (self.factory.make_string(), None, self.factory.make_string()),
            (self.factory.make_string(), self.factory.make_string(), None),
        ]:
            sf = self.TestFinder(*values)
            e = self.assertRaises(ValueError, sf.find)
            self.assertEqual(
                "All of name, version and s_type must have a value", str(e)
            )


class TestDiscoveredSourceBase(base.TestCase):
    class TestDiscoveredSource(DiscoveredSource):
        def __init__(self, filename=None, content=None, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.fake_content = content
            self.fake_filename = filename

        def populate_archive(self, temp_dir, tar):
            self.temp_dir = temp_dir
            _, fake_file = tempfile.mkstemp(dir=temp_dir)
            with open(fake_file, 'wb') as fake_file_fd:
                fake_file_fd.write(self.fake_content)
            tar.add(fake_file, arcname=self.fake_filename)

    def test_make_archive_yields_fileobj_to_tar(self):
        filename = self.factory.make_string('filename')
        content = self.factory.make_bytes('content')
        tds = self.TestDiscoveredSource(filename, content, urls=[])

        with tds.make_archive() as fd:
            with tarfile.open(mode='r', fileobj=fd) as tar:
                self.expectThat(tar.getnames(), SameMembers([filename]))
                self.expectThat(
                    tar.extractfile(filename).read(), Equals(content)
                )

    def test_make_archive_cleans_up_temp_dir_after_exit(self):
        filename = self.factory.make_string('filename')
        content = self.factory.make_bytes('content')
        tds = self.TestDiscoveredSource(filename, content, urls=[])
        with tds.make_archive() as _:
            self.assertThat(tds.temp_dir, DirExists())
        self.assertThat(tds.temp_dir, Not(DirExists()))

    def test_download_file_streams_file_to_target(self):
        tmpdir = self.useFixture(fixtures.TempDir()).path
        target = self.factory.make_string('filename')
        url = self.factory.make_url()
        content = self.factory.make_bytes('content')
        response = requests.Response()
        response.status_code = requests.codes.ok
        response.raw = BytesIO(content)
        get = self.patch(requests, 'get')
        get.return_value = response

        f = self.factory.make_string('filename')
        c = self.factory.make_bytes('content')
        tds = self.TestDiscoveredSource(f, c, urls=[])
        returned_path = tds.download_file(tmpdir, target, url)

        expected_path = pathlib.Path(tmpdir) / target
        get.assert_called_once_with(url, stream=True, timeout=30)
        self.assertThat(expected_path, Equals(returned_path))
        # Cannot use matchers.FileContains because it returns decoded
        # strings...damn.
        with open(returned_path, 'rb') as fd:
            self.assertThat(fd.read(), Equals(content))

    def test_download_file_raises_on_http_errors(self):
        tmpdir = self.useFixture(fixtures.TempDir()).path
        target = self.factory.make_string('filename')
        url = self.factory.make_url()
        response = requests.Response()
        response.status_code = requests.codes.not_found
        response.raw = BytesIO(b'')
        get = self.patch(requests, 'get')
        get.return_value = response

        tds = self.TestDiscoveredSource(urls=[])
        exc = self.assertRaises(
            exceptions.DownloadError,
            tds.download_file,
            tmpdir,
            target,
            url,
        )
        self.assertEqual(response.status_code, exc.status_code)

        expected_path = pathlib.Path(tmpdir) / target
        get.assert_called_once_with(url, stream=True, timeout=30)
        self.assertFalse(expected_path.exists())

    def test_filter_tarinfo(self):
        tarinfo = tarfile.TarInfo()
        tarinfo.uid = tarinfo.gid = self.factory.randint(0, 100)
        tarinfo.uname = tarinfo.gname = self.factory.make_string()
        filename = self.factory.make_string('filename')
        content = self.factory.make_bytes('content')
        tds = self.TestDiscoveredSource(filename, content, urls=[])
        tarinfo = tds.reset_tarinfo(tarinfo)
        self.expectThat(tarinfo.uid, Equals(0))
        self.expectThat(tarinfo.gid, Equals(0))
        self.expectThat(tarinfo.uname, Equals('root'))
        self.expectThat(tarinfo.gname, Equals('root'))

    def test_remote_url_is_archive(self):
        class URLDiscoveredSource(DiscoveredSource):
            make_archive = DiscoveredSource.remote_url_is_archive
            populate_archive = lambda: None  # noqa: E731

        tmpdir = self.useFixture(fixtures.TempDir()).path

        # Make a fake tar file
        content = self.factory.make_bytes('content')
        _, fake_file = tempfile.mkstemp(dir=tmpdir)
        with open(fake_file, 'wb') as fake_file_fd:
            fake_file_fd.write(content)

        # Patch out download_file to return the fake file:
        ds = URLDiscoveredSource(urls=[self.factory.make_url()])
        download_file = self.patch(ds, 'download_file')
        download_file.return_value = pathlib.Path(fake_file)

        # Call make_archive to fetch the fake file:
        _, tar_file_name = tempfile.mkstemp(dir=tmpdir)
        with ds.make_archive() as tarfile_fd:
            with open(tar_file_name, 'wb') as f:
                shutil.copyfileobj(tarfile_fd, f)

        # Test that the copied file contains the fake downloaded content.
        self.assertThat(tar_file_name, FileContains(content.decode()))



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

"""Factory functions and classes for tests."""

import hashlib
import random
import string
from datetime import datetime, timezone
from itertools import count, islice, repeat
from unittest.mock import sentinel


class Factory:
    """Class that defines helpers that make things for you."""

    randint = random.randint
    random_choice = random.choice
    random_sample = random.sample
    random_ascii = map(random_choice, repeat(string.ascii_letters))
    random_letters = map(
        random_choice, repeat(string.ascii_letters + string.digits)
    )
    random_sentences = map(random_choice, repeat(string.printable))
    random_whitespace = map(random_choice, repeat(string.whitespace))

    # Marker to instruct factory methods to create a random object from
    # parameters.
    RANDOM = sentinel.RANDOM

    def __init__(self):
        """Initialize a monotonic ID generator.

        For use when making database items.  This will allow for every
        factory-created item in the test DB to have a globally-unique ID
        value, so that comparisons between objects won't have any
        unintentional overlap, but sorting and what-have-you will still
        behave as expected.

        To use, simply pass it in as the `id` kwarg when creating a model
        object, e.g,:
            role_id = next(self.next_id)
            models.Role(id=role_id, name=name, description=desc, project=proj)
        """
        self.next_id = count(1)

    def randomize_case(self, inp):
        return "".join(map(self.random_choice, zip(inp.lower(), inp.upper())))

    def make_string(self, prefix="", size=10, charset=None):
        if charset is None:
            charset = self.random_letters
        return prefix + "".join(islice(charset, size))

    def make_bytes(self, prefix="", size=10, charset=None):
        return self.make_string(prefix, size, charset).encode('utf-8')

    def make_digest(self):
        string = self.make_string().encode("ascii")
        return "sha256:" + hashlib.sha256(string).hexdigest()

    def make_semver(self, octets=3, prerelease=None, build=None, extra=None):
        """Generate a valid (or invalid, if you prefer) semantic version."""
        # TODO(nic): a way to generate a semver that is greater than / less
        #  than another semver, without a needlessly fidgety interface
        semver = ".".join(str(self.randint(0, 9)) for _ in range(octets))
        if prerelease is not None:
            semver += f"-{prerelease}"
        if build is not None:
            semver += f"+{build}"
        if extra is not None:
            semver += str(extra)
        return semver

    def make_domainname(self, prefix=None):
        # https://en.wikipedia.org/wiki/List_of_Internet_top-level_domains#Original_top-level_domains
        tlds = ("com", "org", "net", "int", "edu", "gov", "mil", "arpa")
        if prefix is None:
            prefix = self.make_string("site")
        return "{}.{}".format(prefix, self.random_choice(tlds))

    def make_url(self, scheme="http"):
        return "{}://{}/{}".format(
            scheme, self.make_domainname(), self.make_string("path")
        )

    def pick_enum(self, enum, but_not=frozenset()):
        return self.random_choice(
            [value for value in enum if value not in but_not]
        )

    def pick_bool(self):
        return self.random_choice([True, False])

    def make_datetime(self, from_utime=None):
        datetimes = (
            # Apollo 11 landing date
            datetime(1969, 7, 20, 20, 17, 40, tzinfo=timezone.utc),
            # Mars Pathfinder landing date
            datetime(1997, 7, 4, 16, 56, 55, tzinfo=timezone.utc),
            # New Horizons Pluto flyby date
            datetime(2015, 7, 14, 11, 49, 57, tzinfo=timezone.utc),
            # Venera 5 landing date
            datetime(1969, 5, 16, 6, 54, tzinfo=timezone.utc),
            # Huygens probe landing date
            datetime(2005, 1, 14, 12, 43, tzinfo=timezone.utc),
            # time_t == 0x7FFFFFFF
            datetime(2038, 1, 18, 19, 14, 7, tzinfo=timezone.utc),
        )
        if from_utime is not None:
            return datetime.fromtimestamp(from_utime, tz=timezone.utc)
        return self.random_choice(datetimes)


# Factory is a singleton.
factory = Factory()
